#!/usr/bin/perl -w

###########################################################################
#
# Licensed Materials - Property of IBM
# 5724-T10
#
# (C) Copyright IBM Corp. 2003,2008. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
###########################################################################

#
# $Header: /nfs/kl/gateways/da/CVSROOT/modules/BWAAA/scripts/remediation_ErcAAA.pl,v 1.4 2008/07/02 03:08:29 khalidm Exp $
#
#
# remediation_ErcAAA.pl
#
#         This script is used to implement data remediation for Ericsson AAA
#
# Usage:  remediation_ErcAAA.pl <config filename> <DC filename>
#                               <preceding timestamp> <preceding status>
#                               <DC interval> <report interval> <outputfile format> <outputfile path>
#
# Example:
#         remediation_ErcAAA.pl erc.aaa.sc.oid.cfg DC.LA.1.200203060030.AAApm
#                               200203060015 s 15 60 CHICO.1:AAApm /dest

use strict;
use Time::Local;
use File::Basename;

usage("Invalid number of arguments") if ($#ARGV != 9);

my $maxUnsigned32 = 2**32;
my $maxUnsigned31 = 2**31;

my $minReportInterval = 60;
my $tstampLength = 14;

my $parentEntitySC = "AAA_SrvcController_Erc";
my $ruleBasedAuthentication = "radiusAuthServUpTime";
my $ruleBasedAccounting = "radiusAcctServUpTime";

my $parentEntityProxy = "AAA_ProxyServer_Erc";
my $ruleBasedProxy1 = "radiusAuthClientTotPktSent";
my $ruleBasedProxy2 = "radiusAuthClientTotAccReq";
my $ruleBasedProxy3 = "radiusAuthClientTotResp";
my $ruleBasedProxy4 = "radiusAcctClientTotPktSent";
my $ruleBasedProxy5 = "radiusAcctClientTotReq";
my $ruleBasedProxy6 = "radiusAcctClientTotResp";

my $gzFilePostfix = "gz";
my $gzipCommand = "gzip";
my $gunzipCommand = "gunzip";

my $posServerIdFileName = 3;

## Old da6.0 file format DC.LA.1.200705170215.AAAProxypm
## 			 ^  ^  ^ ^            ^
## Position:             1  2  3 4            5
## New da7.0 file format 200705172115.AAAProxypm
##                       ^            ^
##                       1            2
## As you can see from above, the position of timestamp in da7.0 filename is at 1. 
## Thus, position of timestamp filename is changed from 4 to 1.
## my $posTstampFileName = 4;
my $posTstampFileName = 1;

my $posServerFileName = 2;
my $scFileName = "AAApm";
my $proxyFileName = "AAAProxypm";
my $delimiterFileField = '\.';
my $delimiterDirField = '/';

my $minConfigField = 12;
my $posOidConfig = 1;
my $posOidNameConfig = 2;
my $posAccumulatingConfig = 7;
my $posRuleBasedConfig = 8;
my $posMibFileConfig = 9;
my $posEntityConfig = 10;
my $posLocalIdConfig = 11;
my $posDataTypeConfig = 12;
my $delimiterConfig = '\~';


## da6.0 sample data: 1.3.6.1.2.1.67.1.1.1.1.2|radiusAuthServUpTime|0|0:08:50.00|1179337503
## da7.0 sample data: 1.3.6.1.2.1.67.1.1.1.1.2|radiusAuthServUpTime|0|0d  0:58:03.00|
## As you can see from above, the da7.0 data has one less column, thus the minimum dc field is reduced
## from 5 to 4. This variable is used below to check the validity of the input data,
## make sure it has the at least minimum number of columns.
## my $minDcField = 5;
my $minDcField = 4;

my $posTstampLineDc = 1;
my $posTstampColDc = 1;
my $posOidNameDc = 2;
my $posInstDc = 3;
my $posValueDc = 4;
my $posTstampDc = 5;
my $delimiterDc = '\|';

my $success = "s";
my $fail = "x";
my $unknown = "u";
my $true = "true";
my $false = "false";
my $resetTrue = "restarted";
my $resetFalse = "not-restarted";
my $resetNoRule = "unknown-rule";
my $resetUnknown = "unknown-previous";
my $compareTrue = "greater-or-equal-to";
my $compareFalse = "less-than";

my $unsigned32Type = "UNSIGNED32";
my $unsigned31Type = "UNSIGNED31";
my $timeType = "TIMETICKS";
my $stringType = "STRING";
my $gaugeType = "GAUGE32";
my $integerType = "INTEGER";
my $numberLiteral = '^\d+$';
my @timeCommonDelimiter = ("\:", "\:", "\.");


## TimeTicks format expected in da6.0 => D day HH:MM:SS.ss
my $dLiteral = "d";
my $dayLiteral = "day";
my $daysLiteral = "days,";
#my $timeLiteral = '^(\d*\s*day(s,)?\s*)?\d+\:\d\d\:\d\d\.\d\d$';
my $timeLiteral = '^(\d*\s*(d|day)(s,)?\s*)?\d+\:\d\d\:\d\d\.\d\d$';
my $oidKey = "OID";
my $accumulatingKey = "ACCUMULATING";
my $ruleBasedKey = "RULEBASED";
my $mibFileKey = "MIBFILE";
my $entityKey = "ENTITY";
my $localIdKey = "LOCALID";
my $dataTypeKey = "DATATYPE";
my $instanceListExpr = "INSTANCELIST";
my $memberKey = "MEMBER";
my $instanceKey = "INSTANCE";
my $resetKey = "RESTART";
my $currValueKey = "CURRVALUE";
my $currTstampKey = "CURRTSTAMP";
my $prevValueKey = "PREVVALUE";
my $prevTstampKey = "PREVTSTAMP";

my $refOidNameKey = "OIDNAME";
my $instanceCountKey = "TOTALINSTANCES";
my $collectMinKey = "CollectionMins";
my $delimiterInstanceExpr = ',';

my ($dcInterval, $reportInterval, $startTime, $numInterval);
my (%myData, %myEntity, $dataDirName, $currDcBaseName);
my ($configFile, $currDcFile);

my ($currDcOrigFile, $prevDcOrigFile, $prevDcFile);
my (@configResult, @currDcResult, @prevDcResult);
my ($currDcGzFlag, $currTstampDc, $currTstamp, $currOidName, $currInstanceKey);
my ($prevDcGzFlag, $prevTstampDc, $prevTstamp, $prevStatus, $prevOidName);
my ($gzPostfix, $gzPos, $lineNum, @lineWords, @fldWords);
my ($collectMinutes);
my ($realInterval, $minDc, $hourDc, $ydayDc);
my ($statusAuthentication, $statusAccounting);
my ($statusProxy, $statusProxy1, $statusProxy2, $statusProxy3, $statusProxy4, $statusProxy5, $statusProxy6);
my ($ruleBasedMibFile, $serverFileName);

# record type data structure to handle all information
#
# %myData = {
#   radiusAuthServUpTime => {
#     $oidKey => "1.3.6.1.2.1.67.1.1.1.1.2",
#     $accumulatingKey => "N",
#     $ruleBasedKey => "Y",
#     $mibFileKey => "radiusAuthServer",
#     $entityKey => "AAA_SrvcController_Erc",
#     $localIdKey => "N",
#     $dataTypeKey => "TIMETICKS",
#     $instanceListExpr => "0" 
#     $memberKey => {
#        0 => {
#        $currValueKey => "39days, 15:39:08.00",
#        $currTstampKey => "1021661101",
#        $prevValueKey => "39days, 15:09:08.00",
#        $prevTstampKey => "1021660202",},
#     },
#   },
#
#   radiusAccServerAddress => {
#     $oidKey => "1.3.6.1.2.1.67.2.2.1.1.3.1.2",
#     $accumulatingKey => "N",
#     $ruleBasedKey => "N",
#     $mibFileKey => "radiusAccClient",
#     $entityKey => "AAA_ConnAcctServer_Erc",
#     $localIdKey => "Y",
#     $dataTypeKey => "STRING",
#     $instanceListExpr => "1,2" 
#     $memberKey => {
#       1 => {
#        $currValueKey => "1.2.3.4",
#        $currTstampKey => "1021661102",},
#       2 => {
#        $currValueKey => "2.3.4.5",
#        $currTstampKey => "1021661103",},
#     },
#   },
#
#   radiusAuthServTotAccAccept => {
#     $oidKey => "1.3.6.1.2.1.67.1.1.1.1.8",
#     $accumulatingKey => "Y",
#     $ruleBasedKey => "N",
#     $mibFileKey => "radiusAuthServer",
#     $entityKey => "AAA_SrvcController_Erc",
#     $localIdKey => "N",
#     $dataTypeKey => "UNSIGNED32",
#     $instanceListExpr => "0" 
#     $memberKey => {
#       0 => {
#        $resetKey => "not-restarted";
#        $currValueKey => "31046",
#        $currTstampKey => "1021661104",
#        $prevValueKey => "31040",
#        $prevTstampKey => "1021660204",},
#     },
#   },
# ...
# };

# record type data structure to handle all auxiliary information
#
# %myEntity = {
#   AAA_SrvcController_Erc => {
#     $refOidNameKey => "radiusAuthServUpTime",
#     $instanceCountKey => 1,
#     $instanceKey => {
#       0 => 0,
#     },
#     $memberKey => [
#       {$mibFileKey => "radiusAuthServer",
#        $collectMinKey => "15",},
#       {$mibFileKey => "radiusAccServer",
#        $collectMinKey => "",},
#     ],
#   },
#
#   AAA_ConnAuthServer_Erc => {
#     $refOidNameKey => "radiusAuthServerAddress",
#     $instanceCountKey => 1,
#     $instanceKey => {
#       0 => 0,
#     },
#     $memberKey => [
#       {$mibFileKey => "radiusAuthClient",
#        $collectMinKey => "15",},
#     ],
#   },
#
#   AAA_ConnAcctServer_Erc => {
#     $refOidNameKey => "radiusAccServerAddress",
#     $instanceCountKey => 4,
#     $instanceKey => {
#       1 => 1,
#       2 => 2,
#       3 => 3,
#       4 => 4,
#     },
#     $memberKey => [
#       {$mibFileKey => "radiusAccClient",
#        $collectMinKey => "",},
#     ],
#   },
# ...
# };

$configFile = $ARGV[0];
$currDcOrigFile = $ARGV[1];
$prevTstamp = $ARGV[2];
$prevStatus = $ARGV[3];
$dcInterval = $ARGV[4];
$reportInterval = $ARGV[5];

## FILE_PREFIX  need to be further broken down to retrieve the OutputId of the file
## EXAMPLE FILE HEADER
## ===================
## STARTTIME|200705170200
## ENDTIME|200705170215
## SERVER|2             <-- This is the OutputID
## ENTITY|AAA_AuthClients_Erc
## START
## INSTANCE|1

my($outputFilePrefix)="";
my($outputFileSuffix) = "";
my($outputFileId) = "";
($outputFilePrefix,$outputFileSuffix) = split(/:/,$ARGV[6]);
if ( ($outputFilePrefix eq "" ) || ($outputFileSuffix eq "") )
{
  terminalError("The file output format is wrong.  Illegal filename  output format $ARGV[6]");
}

my(@filePrefixItems)=split(/\./,$outputFilePrefix);
if ( ($outputFileId=$filePrefixItems[1]) eq "")
{
  terminalError("Unable to retrieve Output-File-Id from $ARGV[6]. Example: CHICO.1:AAApm where Output-File-Id is 1.");
}

my($outputFilePath)=$ARGV[7];
if ( $outputFilePath eq "")
{
  terminalError("Invalid output file path $ARGV[7]");
}


my($prevFilePath1)=$ARGV[8];
if ( $prevFilePath1 eq "")
{
  terminalError("Invalid previous file path $ARGV[8]");
}
my($prevFilePath2)=$ARGV[9];
if ( $prevFilePath2 eq "")
{
  terminalError("Invalid previous file path $ARGV[9]");
}
# initialize $startTime and $collectMinutes
$startTime = $prevTstamp;
$collectMinutes = "";

# gunzip $currDcOrigFile when necessary
$gzPos = length($currDcOrigFile);
if (($gzPos = rindex($currDcOrigFile, ".", $gzPos)) >= 0) {
  $gzPostfix = substr($currDcOrigFile, $gzPos + 1);

  if ($gzPostfix eq $gzFilePostfix) {
    gzSysFunction($gunzipCommand, $currDcOrigFile);
    $currDcFile = substr($currDcOrigFile, 0, $gzPos);
    $currDcGzFlag = $true;
  }
  else {
    $currDcFile = $currDcOrigFile;
    $currDcGzFlag = $false;
  }
}
else {
  terminalError("The field separator in filename is wrong.  Illegal filename $currDcOrigFile");
}

usage("Could not open config file $configFile") if (!open(configFileHandle, $configFile));
terminalError("Config file $configFile is empty") if -z $configFile;

usage("Could not open current DC result file $currDcFile") if (!open(currDcFileHandle, $currDcFile));
terminalError("Current DC result file $currDcFile is empty") if -z $currDcFile;

# read all lines from config file
@configResult = <configFileHandle>;
close (configFileHandle);

# read all lines from current data collection result
@currDcResult = <currDcFileHandle>;
close (currDcFileHandle);

if ($currDcGzFlag eq $true) {
  gzSysFunction($gzipCommand, $currDcFile);
}

if (($prevTstamp !~ /$numberLiteral/) ||
    length($prevTstamp) != ($tstampLength )) {
    #length($prevTstamp) != ($tstampLength - 2)) {
  usage("Invalid timestamp $prevTstamp");
}
elsif ($prevStatus ne $success &&
       $prevStatus ne $fail &&
       $prevStatus ne $unknown) {
  usage("Unknown status $prevStatus supplied");
}
elsif ($reportInterval !~ /$numberLiteral/ ||
       $reportInterval < $minReportInterval) {
  usage("Invalid report interval $reportInterval supplied");
}
elsif ($dcInterval !~ /$numberLiteral/ ||
       $dcInterval > $reportInterval) {
  usage("Invalid DC interval $dcInterval supplied");
}

# retrieve directory and base name
$dataDirName = dirname($currDcFile);
$currDcBaseName = basename($currDcFile);

# read configFile for $oidKey, $accumulatingKey, $ruleBasedKey,
# $mibFileKey, $entityKey, $localIdKey, and $dataTypeKey
for ($lineNum = 0; $lineNum < @configResult; $lineNum++) {
  chomp($configResult[$lineNum]);

  # skip comments
  if ($configResult[$lineNum] !~ /^#/) {
    @lineWords = ();
    @lineWords = split ($delimiterConfig, $configResult[$lineNum]);

    if (@lineWords >= $minConfigField) {
      $currOidName = $lineWords[$posOidNameConfig-1];
      $myData{$currOidName}{$oidKey} =
                                 $lineWords[$posOidConfig-1];
      $myData{$currOidName}{$accumulatingKey} =
                                 $lineWords[$posAccumulatingConfig-1];
      $myData{$currOidName}{$ruleBasedKey} =
                                 $lineWords[$posRuleBasedConfig-1];
      $myData{$currOidName}{$mibFileKey} =
                                 $lineWords[$posMibFileConfig-1];
      $myData{$currOidName}{$entityKey} =
                                 $lineWords[$posEntityConfig-1];
      $myData{$currOidName}{$localIdKey} =
                                 $lineWords[$posLocalIdConfig-1];
      $myData{$currOidName}{$dataTypeKey} =
                                 $lineWords[$posDataTypeConfig-1];
    }
    else {
      terminalError("Missing fields at line $lineNum: $configResult[$lineNum] in $configFile");
    }
  }
}

# read $currDcFile for $currTstampDc, $instanceKey,
# $currValueKey, and $currTstampKey

chomp($currDcResult[$posTstampLineDc-1]);
$currTstampDc = parseTimeStamp($currDcResult[$posTstampColDc-1]);

if ($currTstampDc !~ /$numberLiteral/) {
  terminalError("Invalid timestamp $currTstampDc at line $posTstampLineDc of $currDcFile");
}

for ($lineNum = $posTstampLineDc; $lineNum < (@currDcResult-1); $lineNum++) {
  chomp($currDcResult[$lineNum]);

  # skip comments
  if ($currDcResult[$lineNum] !~ /^#/) {
    @lineWords = ();
    @lineWords = split ($delimiterDc, $currDcResult[$lineNum]);

    if (@lineWords >= $minDcField) {
      $currOidName = $lineWords[$posOidNameDc-1];

      $currInstanceKey = $lineWords[$posInstDc-1];
      $myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey} =
                                           $lineWords[$posValueDc-1];
      $myData{$currOidName}{$memberKey}{$currInstanceKey}{$currTstampKey} =
					   $currTstampDc;
                                           #$lineWords[$posTstampDc-1];

      if (defined $myData{$currOidName}{$instanceListExpr}) {
        $myData{$currOidName}{$instanceListExpr}
                                     .= ",".($lineWords[$posInstDc-1]);
      }
      else {
        $myData{$currOidName}{$instanceListExpr} = $lineWords[$posInstDc-1];
      }
    }
    else {
      terminalError("Missing fields at line $lineNum: $currDcResult[$lineNum] in $currDcFile");
    }
  }
}

# data verification for non-accumulating fields
verifyInstances($accumulatingKey, "N");

# validate $prevStatus for its $success
if ($prevStatus eq $success) {
  # resemble $prevDcFile from $currDcFile and $prevTstamp
  @fldWords = ();
  @fldWords = split ($delimiterFileField, $currDcBaseName);

  $currTstamp = $fldWords[$posTstampFileName - 1];

  if (($posTstampFileName - 1) != 0) {
    $prevDcFile = $fldWords[0];
  }
  else {
    $prevDcFile = $prevTstamp;
  }

  for ($lineNum = 1; $lineNum < @fldWords; $lineNum++) {
    if ($lineNum != ($posTstampFileName - 1)) {
      $prevDcFile .= ".".($fldWords[$lineNum]);
    }
    else {
      $prevDcFile .= ".".($prevTstamp);
    }
  }

  my($prevDcFile1) = ($prevFilePath1).($delimiterDirField).($prevDcFile);
  my($prevDcFile2) = ($prevFilePath1).($delimiterDirField).($prevDcFile).".".($gzFilePostfix);
  my($prevDcFile3) = ($prevFilePath2).($delimiterDirField).($prevDcFile);
  my($prevDcFile4) = ($prevFilePath2).($delimiterDirField).($prevDcFile).".".($gzFilePostfix);
  #print "file1=$prevDcFile1\n"; 
  #print "file2=$prevDcFile2\n"; 
  #print "file3=$prevDcFile3\n"; 
  #print "file4=$prevDcFile4\n"; 

  if (-e $prevDcFile1) 
  {
    $prevDcFile=$prevDcFile1;
    $prevDcGzFlag = $false;
  }
  elsif (-e $prevDcFile2)
  {
     gzSysFunction($gunzipCommand, $prevDcFile2);
     $prevDcFile=$prevDcFile1;
     $prevDcGzFlag = $true;
  }
  elsif (-e $prevDcFile3) 
  {
    $prevDcFile=$prevDcFile3;
    $prevDcGzFlag = $false;
  }
  elsif (-e $prevDcFile4)
  {
     gzSysFunction($gunzipCommand, $prevDcFile4);
     $prevDcFile=$prevDcFile3;
     $prevDcGzFlag = $true;
  }
  else {
      terminalError("Previous DC result file does not exist ($prevDcFile)");
  }

  #if (-e $prevDcFile) {
  #  $prevDcOrigFile = $prevDcFile;
  #  $prevDcGzFlag = $false;
  #}
  #else {
  #  $prevDcOrigFile = ($prevDcFile).".".($gzFilePostfix);
  #
  #  if (-e $prevDcOrigFile) {
  #    # gunzip $prevDcOrigFile
  #   gzSysFunction($gunzipCommand, $prevDcOrigFile);
  #    $prevDcGzFlag = $true;
  #  }
  #  else {
  #    terminalError("Previous DC result file does not exist ($prevDcFile or $prevDcOrigFile)");
  #  }
  #}

  # open $prevDcFile
  usage("Could not open previous DC result file $prevDcFile") if (!open(prevDcFileHandle, $prevDcFile));
  terminalError("Previous DC result file $prevDcFile is empty") if -z $prevDcFile;

  # read all lines from previous data collection result
  @prevDcResult = <prevDcFileHandle>;
  close (prevDcFileHandle);

  if ($prevDcGzFlag eq $true) {
    gzSysFunction($gzipCommand, $prevDcFile);
  }

  # find $prevTstampDc
  chomp($prevDcResult[$posTstampLineDc-1]);
  $prevTstampDc = parseTimeStamp($prevDcResult[$posTstampColDc-1]);

  if ($prevTstampDc !~ /$numberLiteral/) {
    terminalError("Invalid timestamp $prevTstampDc at line $posTstampLineDc of $prevDcFile");
  }

  # set $numInterval based on the data collection start time on DA server

  $realInterval = subtractTstamp($currTstampDc, $prevTstampDc);
  $minDc = substr($realInterval, -4, 2);
  $hourDc = substr($realInterval, -6, 2);
  $ydayDc = substr($realInterval, 0, length($realInterval) - 6);

  $numInterval = int(($minDc + 60*($hourDc + 24*$ydayDc))/$dcInterval + 0.5);

  # validate the availability of previous collection period
  if ($prevStatus eq $success &&
      $numInterval > 1 &&
      acrossRepInterval($currTstamp, $prevTstamp) eq $true) {
    $prevStatus = $unknown;
    print STDOUT "$prevTstamp is not immediately preceding collection period of $currTstamp and this extended interval traverses the Prospect Reporting Interval.  No accumulating counters will send to PROSPECT.\n";

    # reset $startTime
    $startTime = resetStartTime($currTstamp, $dcInterval);
  }
}

if ($prevStatus ne $success) {
  # update %myEntity based on local key info
  verifyLocalId();

  # set $collectMinKey in %myEntity for non-accumulating fields related entities
  setCollectMinutes($accumulatingKey, "N", $collectMinutes);

  # if previous period is not succeeded, then build output file
  # only based on non-accumulating fields
  writeOutput($accumulatingKey, "N");
}

#
#
# otherwise, do the following...
elsif ($prevStatus eq $success) {
  # data verification for accumulating fields
  verifyInstances($accumulatingKey, "Y");

  # update %myEntity based on local key info
  verifyLocalId();

  # read $prevDcFile for $prevValueKey and $prevTstampKey
  # only when its current $instanceKey exists
  for ($lineNum = $posTstampLineDc; $lineNum < (@prevDcResult-1); $lineNum++) {
    chomp($prevDcResult[$lineNum]);

    # skip comments
    if ($prevDcResult[$lineNum] !~ /^#/) {
      @lineWords = ();
      @lineWords = split ($delimiterDc, $prevDcResult[$lineNum]);

      if (@lineWords >= $minDcField) {
        $prevOidName = $lineWords[$posOidNameDc-1];

        $currInstanceKey = $lineWords[$posInstDc-1];

        if (defined $myData{$prevOidName}{$memberKey}{$currInstanceKey}) {

          $myData{$prevOidName}{$memberKey}{$currInstanceKey}{$prevValueKey} =
                                                  $lineWords[$posValueDc-1];
          $myData{$prevOidName}{$memberKey}{$currInstanceKey}{$prevTstampKey} =
						  $prevTstampDc;						
                                                  #$lineWords[$posTstampDc-1];
        }
# comment out to avoid too many output messages 
#        else {
#          print STDOUT "$currValueKey is not available for instance $currInstanceKey of $oidKey $prevOidName. This data will neither be remediated nor be sent to PROSPECT.\n";
#        }
      }
      else {
        terminalError("Missing fields at line $lineNum: $prevDcResult[$lineNum] in $prevDcFile");
      }
    }
  }

  # build $resetKey for all $instanceKey in accumulating fields
  # based on the rule for each $mibFileKey
  # for example,
  #   radiusAuthServer and radiusAuthClient =>
  #     if radiusAuthServUpTime <= the number of minutes since the previous
  #        data was collect, then all instances of radiusAuthServer and
  #        radiusAuthClient are considered "restarted";
  #     if no enough information to detect "restarted", it is "unknown-rule";
  #     otherwise, it is "not-restarted".
  #     in the case of "not-restarted", if previous value is missing for this
  #        counter, it will be set to "unknown-previous" and its value will not
  #        be sent to PROSPECT.

  @fldWords = ();
  @fldWords = split ($delimiterFileField, $currDcBaseName);

  $serverFileName = $fldWords[$posServerFileName-1];

  if ($serverFileName eq $scFileName) {
    # build rule for service controller

    $statusAuthentication = calResetStatus($ruleBasedAuthentication);

    if ($statusAuthentication eq $resetNoRule) {
      print STDOUT "Do not know restart status for service controller (authentication) due to no enough information on $ruleBasedAuthentication.  No related accumulating counters will send to PROSPECT.\n";
    }

    $statusAccounting = calResetStatus($ruleBasedAccounting);

    if ($statusAccounting eq $resetNoRule) {
      print STDOUT "Do not know restart status for service controller (accounting) due to no enough information on $ruleBasedAccounting.  No related accumulating counters will send to PROSPECT.\n";
    }

    # apply rule for radiusAuthServer and radiusAuthClient
    # and set $collectMinKey in %myEntity for $ruleBasedMibFile based entities

    if ($statusAuthentication eq $resetFalse) {
      $collectMinutes = $numInterval*$dcInterval;
    }
    elsif ($statusAuthentication eq $resetTrue) {
      $realInterval = decodeTime($myData{$ruleBasedAuthentication}{$memberKey}{$myData{$ruleBasedAuthentication}{$instanceListExpr}}{$currValueKey});

      $minDc = substr($realInterval, -4, 2);
      $hourDc = substr($realInterval, -6, 2);
      $ydayDc = substr($realInterval, 0, length($realInterval) - 6);

      $collectMinutes = $minDc + 60*($hourDc + 24*$ydayDc);
    }
    else {
      $collectMinutes = "";
    }

    $ruleBasedMibFile = "radiusAuthServer";
    applyResetStatus($ruleBasedMibFile, $statusAuthentication);
    setCollectMinutes($mibFileKey, $ruleBasedMibFile, $collectMinutes);

    $ruleBasedMibFile = "radiusAuthClient";
    applyResetStatus($ruleBasedMibFile, $statusAuthentication);
    setCollectMinutes($mibFileKey, $ruleBasedMibFile, $collectMinutes);

    # apply rule for radiusAccServer and radiusAccClient
    # and set $collectMinKey in %myEntity for $ruleBasedMibFile based entities

    if ($statusAccounting eq $resetFalse) {
      $collectMinutes = $numInterval*$dcInterval;
    }
    elsif ($statusAccounting eq $resetTrue) {
      # reset $collectMinutes
      $realInterval =
        decodeTime($myData{$ruleBasedAccounting}{$memberKey}{$myData{$ruleBasedAccounting}{$instanceListExpr}}{$currValueKey});

      $minDc = substr($realInterval, -4, 2);
      $hourDc = substr($realInterval, -6, 2);
      $ydayDc = substr($realInterval, 0, length($realInterval) - 6);

      $collectMinutes = $minDc + 60*($hourDc + 24*$ydayDc);
    }
    else {
      $collectMinutes = "";
    }

    $ruleBasedMibFile = "radiusAccServer";
    applyResetStatus($ruleBasedMibFile, $statusAccounting);
    setCollectMinutes($mibFileKey, $ruleBasedMibFile, $collectMinutes);

    $ruleBasedMibFile = "radiusAccClient";
    applyResetStatus($ruleBasedMibFile, $statusAccounting);
    setCollectMinutes($mibFileKey, $ruleBasedMibFile, $collectMinutes);
  }
  elsif ($serverFileName eq $proxyFileName) {
    # build rule for proxy server

    $statusProxy1 = calResetStatus($ruleBasedProxy1);
    $statusProxy2 = calResetStatus($ruleBasedProxy2);
    $statusProxy3 = calResetStatus($ruleBasedProxy3);
    $statusProxy4 = calResetStatus($ruleBasedProxy4);
    $statusProxy5 = calResetStatus($ruleBasedProxy5);
    $statusProxy6 = calResetStatus($ruleBasedProxy6);

    if ($statusProxy1 eq $resetTrue &&
        $statusProxy2 eq $resetTrue &&
        $statusProxy3 eq $resetTrue &&
        $statusProxy4 eq $resetTrue &&
        $statusProxy5 eq $resetTrue &&
        $statusProxy6 eq $resetTrue) {
      $statusProxy = $resetTrue;
    }
    elsif ($statusProxy1 eq $resetNoRule ||
           $statusProxy2 eq $resetNoRule ||
           $statusProxy3 eq $resetNoRule ||
           $statusProxy4 eq $resetNoRule ||
           $statusProxy5 eq $resetNoRule ||
           $statusProxy6 eq $resetNoRule) {
      $statusProxy = $resetNoRule;
      print STDOUT "Do not know restart status for proxy server due to no enough information on $ruleBasedProxy1, $ruleBasedProxy2, $ruleBasedProxy3, $ruleBasedProxy4, $ruleBasedProxy5, and $ruleBasedProxy6.  No accumulating counters will send to PROSPECT.\n";
    }
    else {
      $statusProxy = $resetFalse;
    }

    # apply rule for BW-radius
    # and set $collectMinKey in %myEntity for $ruleBasedMibFile based entities

    if ($statusProxy eq $resetFalse) {
      $collectMinutes = $numInterval*$dcInterval;
    }
    else {
      $collectMinutes = "";
    }

    $ruleBasedMibFile = "BW-radius";
    applyResetStatus($ruleBasedMibFile, $statusProxy);
    setCollectMinutes($mibFileKey, $ruleBasedMibFile, $collectMinutes);
  }
  else {
    terminalError("Remediation is not for $serverFileName.  Illegal filename $currDcOrigFile");
  }

  # based on resetKey, loop through accumulating fields
  # if $resetKey = $resetFalse, then remediation will be done
  foreach $currOidName (keys %myData) {
    if ($myData{$currOidName}{$accumulatingKey} eq "Y") {
      foreach $currInstanceKey (sort keys %{$myData{$currOidName}{$memberKey}}) {
        if (defined $myData{$currOidName}{$memberKey}{$currInstanceKey}{$resetKey} && $myData{$currOidName}{$memberKey}{$currInstanceKey}{$resetKey} ne "") {
          if ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$resetKey} eq
              $resetFalse) {
            if (isNumber($myData{$currOidName}{$dataTypeKey}) eq $true) {
              if (isValidNumber($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}, $myData{$currOidName}{$dataTypeKey}) eq $false) {
                terminalError("Invalid $myData{$currOidName}{$dataTypeKey} in $currValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}) for instance $currInstanceKey of $currOidName");
              }
              elsif (isValidNumber($myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}, $myData{$currOidName}{$dataTypeKey}) eq $false) {
                terminalError("Invalid $myData{$currOidName}{$dataTypeKey} in $prevValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}) for instance $currInstanceKey of $currOidName");
              }
              else {
                $myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey} = subtractNumber($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}, $myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}, $myData{$currOidName}{$dataTypeKey});
              }
            }
            elsif ($myData{$currOidName}{$dataTypeKey} eq $timeType) {
              if ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey} !~ /$timeLiteral/) {
                terminalError("Invalid $timeType in $currValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}) for instance $currInstanceKey of $currOidName");
              }
              elsif ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey} !~ /$timeLiteral/) {
                terminalError("Invalid $timeType in $prevValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}) for instance $currInstanceKey of $currOidName");
              }
              else {
                if (compareTime($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}, $myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}) eq $compareTrue) {
                  $myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey} = subtractTime($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}, $myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey});
                }
                else {
                  print STDOUT "$prevValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$prevValueKey}) is greater than $currValueKey ($myData{$currOidName}{$memberKey}{$currInstanceKey}{$currValueKey}) so no remediation will be done for instance $currInstanceKey of $currOidName.\n";
                }
              }
            }
            else {
              terminalError("$accumulatingKey should set to N for $currOidName because it is a $myData{$currOidName}{$dataTypeKey} data type");
            }
          }
        }
        else {
          terminalError("$resetKey undefined for instance $currInstanceKey of $currOidName.  Make sure the OID entry in the config file with a valid $mibFileKey");
        }
      }
    }
  }

  # build output file
  writeOutput();
}

exit(0);

sub verifyInstances {
  # read instance counter from %myData
  # verify the same number of instances and the same instance numbers
  # within the same Entity
  my ($lookField, $lookValue);
  my ($i, $oidName);
  my ($currentInstanceKey, $firstAdditional);
  my (@tempInstanceArray, %tempInstanceHash);

  ($lookField, $lookValue) = @_;

  # set $instanceCountKey and instanceKey
  foreach $oidName (keys %myData) {
    if (defined $myData{$oidName}{$lookField} &&
        $myData{$oidName}{$lookField} ne "") {
      if ($myData{$oidName}{$lookField} eq $lookValue &&
          (defined $myData{$oidName}{$instanceListExpr})) {
        @tempInstanceArray = ();
        @tempInstanceArray = sort split ($delimiterInstanceExpr,
                                         $myData{$oidName}{$instanceListExpr});

        if ((not defined $myEntity{$myData{$oidName}{$entityKey}}) ||
            (not defined $myEntity{$myData{$oidName}{$entityKey}}{$instanceKey})) {
          $myEntity{$myData{$oidName}{$entityKey}}{$refOidNameKey} = $oidName;
          $myEntity{$myData{$oidName}{$entityKey}}{$instanceCountKey} =
                                                    $#tempInstanceArray + 1;
          for ($i = 0; $i < $myEntity{$myData{$oidName}{$entityKey}}{$instanceCountKey}; $i++) {
            $myEntity{$myData{$oidName}{$entityKey}}{$instanceKey}{$tempInstanceArray[$i]} = $tempInstanceArray[$i];
          }
        }
        # postpone update based on local key OID info to verifyLocalId()
        # add additional instance from other OID info
        elsif (defined $myEntity{$myData{$oidName}{$entityKey}} &&
               defined $myEntity{$myData{$oidName}{$entityKey}}{$instanceKey} &&
               $myData{$oidName}{$localIdKey} ne "Y") {
# comment out to avoid too many output messages 
          # warning missing instance for current $oidName
#          for ($i = 0; $i <= $#tempInstanceArray; $i++) {
#            $tempInstanceHash{$tempInstanceArray[$i]} = $tempInstanceArray[$i];
#          }
  
#          foreach $currentInstanceKey (sort keys %{$myEntity{$myData{$oidName}{$entityKey}}{$instanceKey}}) {
#            if (not defined $tempInstanceHash{$myEntity{$myData{$oidName}{$entityKey}}{$instanceKey}{$currentInstanceKey}}) {
#              print STDOUT "The instance ($currentInstanceKey) for \'$myEntity{$myData{$oidName}{$entityKey}}{$refOidNameKey}\' does not appear for $oidName in $entityKey $myData{$oidName}{$entityKey}.\n";
#            }
#          }
  
          # merge data from @tempInstanceArray into $instanceKey of %myEntity
          # to find max set of instance keys
          $firstAdditional = $true;
          for ($i = 0; $i <= $#tempInstanceArray; $i++) {
            if (not defined $myEntity{$myData{$oidName}{$entityKey}}{$instanceKey}{$tempInstanceArray[$i]}) {
# comment out to avoid too many output messages 
              # warning missing instance for previous stored in %myEntity
#              print STDOUT "The instance ($tempInstanceArray[$i]) for $oidName does not appear for \'$myEntity{$myData{$oidName}{$entityKey}}{$refOidNameKey}\' in $entityKey $myData{$oidName}{$entityKey}.\n";

              if ($firstAdditional eq $true) {
                $myEntity{$myData{$oidName}{$entityKey}}{$refOidNameKey}
                                                             .= " ".($oidName);
                $firstAdditional = $false;
              }
  
              $myEntity{$myData{$oidName}{$entityKey}}{$instanceCountKey}++;
              $myEntity{$myData{$oidName}{$entityKey}}{$instanceKey}{$tempInstanceArray[$i]} = $tempInstanceArray[$i];
            }
          }
        }
      }
    }
    else {
      terminalError("$lookField undefined for $oidName in $configFile.  Make sure the OID entry in the config file with a valid $lookField flag");
    }
  }
}

sub verifyLocalId {
  my ($i, $entity, $oidName);
  my ($currentInstanceKey, $excludeLocalList);
  my (@tempInstanceArray, %tempInstanceHash);

  foreach $entity (keys %myEntity) {
    foreach $oidName (keys %myData) {
      if ($myData{$oidName}{$entityKey} eq $entity &&
          $myData{$oidName}{$localIdKey} eq "Y") {
        # no local Key info so no data will be sent to PROSPECT
        # for this entity
        if (not defined $myData{$oidName}{$instanceListExpr}) {
          print STDOUT "There is no instance defined for $localIdKey $oidName in $entityKey $entity.  No data will be sent to PROSPECT for this entity.\n";

          if (defined $myEntity{$entity}{$instanceKey}) {
            # clean up $instanceKey of %myEntity
            delete $myEntity{$entity}{$instanceKey};
  
            $myEntity{$entity}{$refOidNameKey} = $oidName;
            $myEntity{$entity}{$instanceCountKey} = 0;
          }
        }
        # replace with local key OID info
        elsif (defined $myEntity{$entity}{$instanceKey} &&
               $oidName ne $myEntity{$entity}{$refOidNameKey}) {
          @tempInstanceArray = ();
          @tempInstanceArray = sort split ($delimiterInstanceExpr,
                                           $myData{$oidName}{$instanceListExpr});

          # warning missing instance for local key
          for ($i = 0; $i <= $#tempInstanceArray; $i++) {
            $tempInstanceHash{$tempInstanceArray[$i]} = $tempInstanceArray[$i];
          }
  
          foreach $currentInstanceKey (sort keys %{$myEntity{$entity}{$instanceKey}}) {
            if (not defined $tempInstanceHash{$myEntity{$entity}{$instanceKey}{$currentInstanceKey}}) {
              $excludeLocalList = $myEntity{$entity}{$refOidNameKey};
              $excludeLocalList =~ s/$oidName//;
              $excludeLocalList =~ s/^ *//;
              $excludeLocalList =~ s/ *$//;

# comment out to avoid too many output messages 
#              print STDOUT "The instance ($currentInstanceKey) for \'$excludeLocalList\' does not appear for $localIdKey $oidName in $entityKey $entity.  Data will not send to PROSPECT for this instance.\n";
            }
          }
  
          # clean up $instanceKey of %myEntity
          delete $myEntity{$entity}{$instanceKey};
  
          # copy data from @tempInstanceArray into $instanceKey of %myEntity
          # to replace all previous instance keys for local key requirements
          $myEntity{$entity}{$refOidNameKey} = $oidName;
          $myEntity{$entity}{$instanceCountKey} = $#tempInstanceArray + 1;
  
          for ($i = 0; $i < $myEntity{$entity}{$instanceCountKey}; $i++) {
            $myEntity{$entity}{$instanceKey}{$tempInstanceArray[$i]} = $tempInstanceArray[$i];
          }
        }
      }
    }
  }
}

sub searchCollectMinutes {
  my ($entity, $mibFile);
  my ($i);

  ($entity, $mibFile) = @_;

  for ($i = 0; $i <= $#{$myEntity{$entity}{$memberKey}}; $i++) {
    if ($mibFile eq $myEntity{$entity}{$memberKey}[$i]{$mibFileKey}) {
      return $true;
    }
  }
  return $false;
}

sub setCollectMinutes {
  my ($lookField, $lookValue, $setValue);
  my ($i, $oidName, $isExist);

  ($lookField, $lookValue, $setValue) = @_;

  # set $memberKey
  foreach $oidName (keys %myData) {
    if (defined $myData{$oidName}{$lookField} &&
        $myData{$oidName}{$lookField} ne "") {
      if ($myData{$oidName}{$lookField} eq $lookValue &&
          defined $myData{$oidName}{$instanceListExpr}) {
        if ($#{$myEntity{$myData{$oidName}{$entityKey}}{$memberKey}} != -1) {
          $isExist = searchCollectMinutes($myData{$oidName}{$entityKey}, $myData{$oidName}{$mibFileKey});

          if ($isExist eq $true) {
            for ($i = 0; $i <= $#{$myEntity{$myData{$oidName}{$entityKey}}{$memberKey}}; $i++) {
              if ($myData{$oidName}{$mibFileKey} eq $myEntity{$myData{$oidName}{$entityKey}}{$memberKey}[$i]{$mibFileKey}) {
                # reset $collectMinKey
                $myEntity{$myData{$oidName}{$entityKey}}{$memberKey}[$i]{$collectMinKey} = $setValue;
                last;
              }
            }
          }
          else {
            push @{$myEntity{$myData{$oidName}{$entityKey}}{$memberKey}},
                            {(
                              $mibFileKey => $myData{$oidName}{$mibFileKey},
                              $collectMinKey => $setValue,
                            )};
          }
        }
        else {
          push @{$myEntity{$myData{$oidName}{$entityKey}}{$memberKey}},
                            {(
                              $mibFileKey => $myData{$oidName}{$mibFileKey},
                              $collectMinKey => $setValue,
                            )};
        }
      }
    }
    else {
      terminalError("$lookField undefined for $oidName in $configFile.  Make sure the OID entry in the config file with a valid $lookField flag");
    }
  }
}

sub calResetStatus {
  my $oidName = $_[0];
  my ($isReset, $prevCalField, $currCalField, $referenceValue);

  if (exists $myData{$oidName}) {
    if (defined $myData{$oidName}{$instanceListExpr} &&
        index($myData{$oidName}{$instanceListExpr}, ",", -1) == -1) {
      if (isNumber($myData{$oidName}{$dataTypeKey}) eq $true) {
        $prevCalField = $prevValueKey;
        $currCalField = $currValueKey;
        $referenceValue = "";
      }
      elsif ($myData{$oidName}{$dataTypeKey} eq $timeType) {
	#print "oidName = $oidName is $timeType\n";
        $prevCalField = $prevTstampKey;
        $currCalField = $currTstampKey;
        $referenceValue = $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currValueKey};
	#print "prevCalField=$prevCalField, currCalField=$currCalField, referenceValue=$referenceValue\n";
      }
      else {
        terminalError("$ruleBasedKey should not set to $myData{$oidName}{$dataTypeKey} for $oidName");
      }
      #print "Long var=".$myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField}."\n";
      if ((not defined $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField}) ||
          $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField} eq "" ||
          (not defined $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currCalField}) ||
          $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currCalField} eq "" ||
          ($myData{$oidName}{$dataTypeKey} eq $timeType &&
           ((not defined $referenceValue) || $referenceValue eq ""))) {
        $isReset = $resetNoRule;
      }
      elsif ($myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currCalField} !~
                                                   /$numberLiteral/) {
        terminalError("Invalid $currCalField ($myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currCalField}) for $oidName");
      }
      elsif ($myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField} !~
                                                   /$numberLiteral/) {
        terminalError("Invalid $prevCalField ($myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField}) for $oidName");
      }
      elsif ($myData{$oidName}{$dataTypeKey} eq $timeType &&
             $referenceValue !~ /$timeLiteral/) {
        terminalError("Invalid $timeType type value ($referenceValue) for $oidName");
      }
      else {
        $isReset = resetStatus($myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$currCalField}, $myData{$oidName}{$memberKey}{$myData{$oidName}{$instanceListExpr}}{$prevCalField}, $referenceValue);
        print STDOUT "$oidName is $isReset.\n";
      }
    }
    else {
      $isReset = $resetNoRule;
    }
  }
  else {
    $isReset = $resetNoRule;
  }

  return $isReset;
}

sub decodeTime {
  my $timeValue = $_[0];
  my ($sec100Value, $secValue, $minValue, $hourValue, $ydayValue);
  my ($i, $pos, $startIndex, @dateField, $timeString);
  my @timeDelimiter = @timeCommonDelimiter;

  if ($timeValue !~ /$timeLiteral/) {
    terminalError("Invalid $timeType type value ($timeValue).  It could not be decoded");
  }
  elsif ($timeValue =~ /$daysLiteral/) {
    unshift @timeDelimiter, $daysLiteral;
  }
  elsif ($timeValue =~ /$dayLiteral/) {
    unshift @timeDelimiter, $dayLiteral;
  }
  elsif ($timeValue =~ /$dLiteral/) {
    unshift @timeDelimiter, $dLiteral;
  }

  $pos = -1;
  $startIndex = 0;
  for ($i = 0; $i < @timeDelimiter; $i++) {
    $pos = index($timeValue, $timeDelimiter[$i], $pos);

    if ($pos > 0) {
      if ($i == 0) {
        $dateField[$i] = substr($timeValue, $startIndex, $pos - $startIndex);
      }
      else {
        $dateField[$i] = substr($timeValue,
                                $startIndex + length($timeDelimiter[$i-1]),
                                $pos - ($startIndex+length($timeDelimiter[$i-1])));
      }
      $startIndex = $pos;
    }
    else {
      $dateField[$i] = "";
    }
    $pos += length($timeDelimiter[$i]);
  }

  # decode the last field (in hundredths of a second)
  $dateField[$i] = substr($timeValue,
                          $startIndex + length($timeDelimiter[$i-1]));

  if (($timeValue !~ /$daysLiteral/) && ($timeValue !~ /$dayLiteral/) && ($timeValue !~ /$dLiteral/)) {
    unshift @dateField, 0;
  }

  $dateField[0] =~ s/^ *//;
  $dateField[0] =~ s/ *$//;
  $ydayValue = $dateField[0];

  $dateField[1] =~ s/^ *//;
  $dateField[1] =~ s/ *$//;
  $hourValue = (length($dateField[1]) == 2) ? $dateField[1] : ("0".($dateField[1]));

  $dateField[2] =~ s/^ *//;
  $dateField[2] =~ s/ *$//;
  $minValue = (length($dateField[2]) == 2) ? $dateField[2] : ("0".($dateField[2]));

  $dateField[3] =~ s/^ *//;
  $dateField[3] =~ s/ *$//;
  $secValue = (length($dateField[3]) == 2) ? $dateField[3] : ("0".($dateField[3]));

  $dateField[4] =~ s/^ *//;
  $dateField[4] =~ s/ *$//;
  $sec100Value = (length($dateField[4]) == 2) ? $dateField[4] : (($dateField[4])."0");

  $timeString = ($ydayValue).($hourValue).($minValue).($secValue).($sec100Value);

  return $timeString;
}

sub encodeTime {
  my ($sec100Value, $secValue, $minValue, $hourValue, $ydayValue);
  my ($i, @timeArray, $timeValue);
  my @timeDelimiter = @timeCommonDelimiter;

  ($ydayValue, $hourValue, $minValue, $secValue, $sec100Value) = @_;

  $hourValue = ($hourValue > 9) ? $hourValue : ("0".($hourValue));
  $minValue = ($minValue > 9) ? $minValue : ("0".($minValue));
  $secValue = ($secValue > 9) ? $secValue : ("0".($secValue));
  $sec100Value = ($sec100Value > 9) ? $sec100Value : (($sec100Value)."0");

  @timeArray = ();
  push @timeArray, $hourValue;
  push @timeArray, $minValue;
  push @timeArray, $secValue;
  push @timeArray, $sec100Value;

  if ($ydayValue != 0) {
    unshift @timeArray, $ydayValue;
    if ($ydayValue == 1) {
      unshift @timeDelimiter, $dayLiteral;
    }
    else {
      unshift @timeDelimiter, $daysLiteral;
    }
  }

  $timeValue = "";
  for ($i = 0; $i < @timeDelimiter; $i++) {
    $timeValue .= ($timeArray[$i]).($timeDelimiter[$i]);
  }

  # encode the last field (in hundredths of a second)
  $timeValue .= $timeArray[$i];

  return $timeValue;
}

sub subtractTstamp {
  my ($currDataTstamp, $prevDataTstamp);
  my ($secDiff, $minDiff, $hourDiff, $ydayDiff);

  ($currDataTstamp, $prevDataTstamp) = @_;

  ($secDiff, $minDiff, $hourDiff, $ydayDiff) =
             (gmtime($currDataTstamp - $prevDataTstamp))[0..2,7];

  $hourDiff = ($hourDiff > 9) ? $hourDiff : ("0".($hourDiff));
  $minDiff = ($minDiff > 9) ? $minDiff : ("0".($minDiff));
  $secDiff = ($secDiff > 9) ? $secDiff : ("0".($secDiff));

  return ($ydayDiff).($hourDiff).($minDiff).($secDiff);
}

sub acrossRepInterval {
  my ($currDataTstamp, $prevDataTstamp);

  ($currDataTstamp, $prevDataTstamp) = @_;

  if (substr($currDataTstamp, 0, 10) == substr($prevDataTstamp, 0, 10)) {
    # count daylight saving time
    if (($currDataTstamp > $prevDataTstamp ||
         ($currDataTstamp <= $prevDataTstamp &&
          substr($currDataTstamp, 10, 2) eq "00")) &&
        $numInterval <= $reportInterval/$dcInterval) {
      return $false;
    }
    else {
      return $true;
    }
  }
  elsif (substr($currDataTstamp, 10, 2) eq "00" &&
         $numInterval <= $reportInterval/$dcInterval) {
    return $false;
  }
  else {
    return $true;
  }
}

sub resetStartTime {
  my $secMin = 60;
  my ($oldTimeString, $offsetMinutes);
  my ($sec, $min, $hour, $mday, $mon, $year);
  my ($newTimeString);

  ($oldTimeString, $offsetMinutes) = @_;

  ($sec, $min, $hour, $mday, $mon, $year) =
            (localtime(timelocal(
                                   0,
                                   substr($oldTimeString, 10, 2),
                                   substr($oldTimeString, 8,  2),
                                   substr($oldTimeString, 6,  2),
                                  (substr($oldTimeString, 4,  2) - 1),
                                  (substr($oldTimeString, 0,  4) - 1900)
                                )
                                - $offsetMinutes*$secMin))[0..5];

  $year += 1900;
  $mon++;
  $mon = ($mon > 9) ? $mon : ("0".($mon));
  $mday = ($mday > 9) ? $mday : ("0".($mday));
  $hour = ($hour > 9) ? $hour : ("0".($hour));
  $min = ($min > 9) ? $min : ("0".($min));

  $newTimeString = ($year).($mon).($mday).($hour).($min);
  return $newTimeString;
}

sub resetStatus {
  my $compareValue = $_[2];

  my ($currDataValue, $prevDataValue);
  my ($currDataTstamp, $prevDataTstamp);
  my ($timeDiff, $timeData);

  if ($compareValue eq "") {
    ($currDataValue, $prevDataValue) = ($_[0], $_[1]);

    if ($currDataValue < $prevDataValue) {
      return $resetTrue;
    }
    else {
      return $resetFalse;
    }
  }
  else {
    ($currDataTstamp, $prevDataTstamp) = ($_[0], $_[1]);

    $timeDiff = subtractTstamp($currDataTstamp, $prevDataTstamp);

    $timeData = decodeTime($compareValue);

    if (substr($timeData, 0, length($timeData) - 2) <= $timeDiff) {
      return $resetTrue;
    }
    else {
      return $resetFalse;
    }
  }
}

sub applyResetStatus {
  my ($mibFileName, $resetValue);
  my ($oidName, $currentInstanceKey);

  ($mibFileName, $resetValue) = @_;

  foreach $oidName (keys %myData) {
    if ($myData{$oidName}{$mibFileKey} eq $mibFileName &&
        $myData{$oidName}{$accumulatingKey} eq "Y" &&
        defined $myData{$oidName}{$instanceListExpr}) {
      foreach $currentInstanceKey (sort keys %{$myData{$oidName}{$memberKey}}) {
        if (defined $myData{$oidName}{$memberKey}{$currentInstanceKey}{$prevValueKey} && $myData{$oidName}{$memberKey}{$currentInstanceKey}{$prevValueKey} ne "") {
          $myData{$oidName}{$memberKey}{$currentInstanceKey}{$resetKey} = $resetValue;
        }
        else {
          $myData{$oidName}{$memberKey}{$currentInstanceKey}{$resetKey} = $resetUnknown;
# comment out to avoid too many output messages 
#          print STDOUT "$prevValueKey undefined for instance $currentInstanceKey of $oidName.  There will be no data output for this instance of the accumulating counter.\n";
        }
      }
    }
  }
}

sub compareTime {
  my ($currTimeValue, $prevTimeValue);
  my ($currTimeString, $prevTimeString);

  ($currTimeValue, $prevTimeValue) = @_;

  $currTimeString = decodeTime($currTimeValue);
  $prevTimeString = decodeTime($prevTimeValue);

  if ($currTimeString >= $prevTimeString) {
    return $compareTrue;
  }
  else {
    return $compareFalse;
  }
}

sub subtractTime {
  my ($currTimeValue, $prevTimeValue);
  my ($currTimeString, $prevTimeString);
  my ($sec100Curr, $secCurr, $minCurr, $hourCurr, $ydayCurr);
  my ($sec100Prev, $secPrev, $minPrev, $hourPrev, $ydayPrev);
  my ($sec100Value, $secValue, $minValue, $hourValue, $ydayValue);
  my ($secBorrow, $minBorrow, $hourBorrow, $ydayBorrow);
  my ($resultTimeValue);

  ($currTimeValue, $prevTimeValue) = @_;
  ($secBorrow, $minBorrow, $hourBorrow, $ydayBorrow) = (0,0,0,0);

  $currTimeString = decodeTime($currTimeValue);
  $prevTimeString = decodeTime($prevTimeValue);

  $sec100Curr = substr($currTimeString, -2, 2);
  $secCurr = substr($currTimeString, -4, 2);
  $minCurr = substr($currTimeString, -6, 2);
  $hourCurr = substr($currTimeString, -8, 2);
  $ydayCurr = substr($currTimeString, 0, length($currTimeString) - 8);

  $sec100Prev = substr($prevTimeString, -2, 2);
  $secPrev = substr($prevTimeString, -4, 2);
  $minPrev = substr($prevTimeString, -6, 2);
  $hourPrev = substr($prevTimeString, -8, 2);
  $ydayPrev = substr($prevTimeString, 0, length($prevTimeString) - 8);

  if ($sec100Curr >= $sec100Prev) {
    $sec100Value = $sec100Curr - $sec100Prev;
  }
  else {
    $sec100Value = 100 + $sec100Curr - $sec100Prev;
    $secBorrow++;
  }

  if ($secCurr >= ($secPrev + $secBorrow)) {
    $secValue = $secCurr - ($secPrev + $secBorrow);
  }
  else {
    $secValue = 60 + $secCurr - ($secPrev + $secBorrow);
    $minBorrow++;
  }

  if ($minCurr >= ($minPrev + $minBorrow)) {
    $minValue = $minCurr - ($minPrev + $minBorrow);
  }
  else {
    $minValue = 60 + $minCurr - ($minPrev + $minBorrow);
    $hourBorrow++;
  }

  if ($hourCurr >= ($hourPrev + $hourBorrow)) {
    $hourValue = $hourCurr - ($hourPrev + $hourBorrow);
  }
  else {
    $hourValue = 24 + $hourCurr - ($hourPrev + $hourBorrow);
    $ydayBorrow++;
  }

  if ($ydayCurr >= ($ydayPrev + $ydayBorrow)) {
    $ydayValue = $ydayCurr - ($ydayPrev + $ydayBorrow);
  }
  else {
    terminalError("Due to compareTime is performed before subtractTime, this is not a reachable point.  $timeType in $currValueKey ($currTimeValue) is less than $prevValueKey ($prevTimeValue)");
  }

  $resultTimeValue = encodeTime($ydayValue, $hourValue, $minValue, $secValue, $sec100Value);

  return $resultTimeValue;
}

sub isNumber {
  my $numValue = $_[0];

  if ($numValue eq $unsigned32Type || $numValue eq $unsigned31Type) {
    return $true;
  }
  else {
    return $false;
  }
}

sub isValidNumber {
  my ($numValue, $dataType) = @_;

  if ($numValue !~ /$numberLiteral/) {
    return $false;
  }
  elsif ($dataType eq $unsigned32Type && $numValue < $maxUnsigned32) {
    return $true;
  }
  elsif ($dataType eq $unsigned31Type && $numValue < $maxUnsigned31) {
    return $true;
  }
  else {
    return $false;
  }
}

sub subtractNumber {
  my ($currValue, $prevValue, $dataType);
  my ($returnValue);

  ($currValue, $prevValue, $dataType) = @_;

  if ($currValue >= $prevValue) {
    $returnValue = $currValue - $prevValue;
  }
  elsif ($dataType eq $unsigned32Type) {
    $returnValue = $maxUnsigned32 - $prevValue + $currValue;
  }
  elsif ($dataType eq $unsigned31Type) {
    $returnValue = $maxUnsigned31 - $prevValue + $currValue;
  }

  return $returnValue;
}

sub timeToNumber {
  my $timeValue = $_[0];
  my ($minValue, $hourValue, $ydayValue);
  my ($i, $pos, $startIndex, @dateField, $returnValue);
  my @timeDelimiter = @timeCommonDelimiter;

  if ($timeValue =~ /$daysLiteral/) {
    unshift @timeDelimiter, $daysLiteral;
  }
  elsif ($timeValue =~ /$dayLiteral/) {
    unshift @timeDelimiter, $dayLiteral;
  }
  elsif ($timeValue =~ /$dLiteral/) {
    unshift @timeDelimiter, $dLiteral;
  }

  $pos = -1;
  $startIndex = 0;
  for ($i = 0; $i < @timeDelimiter; $i++) {
    $pos = index($timeValue, $timeDelimiter[$i], $pos);

    if ($pos > 0) {
      if ($i == 0) {
        $dateField[$i] = substr($timeValue, $startIndex, $pos - $startIndex);
      }
      else {
        $dateField[$i] = substr($timeValue,
                                $startIndex + length($timeDelimiter[$i-1]),
                                $pos - ($startIndex+length($timeDelimiter[$i-1])));
      }
      $startIndex = $pos;
    }
    else {
      $dateField[$i] = "";
    }
    $pos += length($timeDelimiter[$i]);
  }

  # retrieve the last field (in hundredths of a second)
  $dateField[$i] = substr($timeValue,
                          $startIndex + length($timeDelimiter[$i-1]));

  if ( ($timeValue =~ /$daysLiteral/) || 
       ($timeValue =~ /$dayLiteral/) ||
       ($timeValue =~ /$dLiteral/) )

  {
    $ydayValue = $dateField[0];
    $hourValue = $dateField[1];
    $minValue = $dateField[2];
  }
  else {
    $ydayValue = 0;
    $hourValue = $dateField[0];
    $minValue = $dateField[1];
  }

  $returnValue = $minValue + 60*($hourValue + 24*$ydayValue);

  return $returnValue;
}

sub epochToDate {
  my $epochValue = $_[0];
  my ($sec, $min, $hour, $mday, $mon, $year);
  my ($timeString);

  ($sec, $min, $hour, $mday, $mon, $year) = (localtime($epochValue))[0..5];

  $year += 1900;
  $mon++;
  $mon = ($mon > 9) ? $mon : ("0".($mon));
  $mday = ($mday > 9) ? $mday : ("0".($mday));
  $hour = ($hour > 9) ? $hour : ("0".($hour));
  $min = ($min > 9) ? $min : ("0".($min));
  $sec = ($sec > 9) ? $sec : ("0".($sec));

  $timeString = ($year).($mon).($mday).($hour).($min).($sec);

  return $timeString;
}

sub writeOutput {
  my ($lookField, $lookValue);
  my ($i, $entity, $oidName, @fieldWords);
  my ($outputId, $outputTime, $outputValue, $outputFile);
  my ($currentInstanceKey);

  if (defined @_) {
    ($lookField, $lookValue) = @_;
  }
  else {
    undef $lookField;
  }

  @fieldWords = ();
  @fieldWords = split ($delimiterFileField, $currDcBaseName);

  $outputTime = $fieldWords[0];


  $outputFile = $outputFilePrefix.".".$outputTime.".".$outputFileSuffix;

  #for ($i = 2; $i < @fieldWords; $i++) {
  #  if ($i != ($posTstampFileName - 1)) {
  #    $outputFile .= ".".($fieldWords[$i]);
  #  }
  #  else {
  #    $outputFile .= ".".($outputTime);
  #  }
  #}

  #$outputFile = ($dataDirName).($delimiterDirField).($outputFile);
  my($realOutputFile) = ($outputFilePath).($delimiterDirField).($outputFile);
  $outputFile = ($outputFilePath).($delimiterDirField).($outputFile).".part";

  # open $outputFile for writing output
  terminal("Could not redirect stdout to $outputFile") if (!open(outputFileHandle, ">$outputFile"));

  if (length($startTime)>12)
  {
	$startTime=substr($startTime,0,12);
  }
 
  if (length($outputTime)>12)
  {
	$outputTime=substr($outputTime,0,12);
  }
  print outputFileHandle "STARTTIME|$startTime\n";
  print outputFileHandle "ENDTIME|$outputTime\n";
  print outputFileHandle "SERVER|$outputFileId\n";
  
  foreach $entity (keys %myEntity) {
    print outputFileHandle "ENTITY|$entity\n";

    if ($entity eq $parentEntitySC || $entity eq $parentEntityProxy) {
      for ($i = 0; $i <= $#{$myEntity{$entity}{$memberKey}}; $i++) {
        print outputFileHandle "$myEntity{$entity}{$memberKey}[$i]{$mibFileKey}_$collectMinKey|$myEntity{$entity}{$memberKey}[$i]{$collectMinKey}|\n";
      }
    }

    foreach $currentInstanceKey (sort keys %{$myEntity{$entity}{$instanceKey}}) {
      print outputFileHandle "START\n";
      print outputFileHandle "INSTANCE|$myEntity{$entity}{$instanceKey}{$currentInstanceKey}\n";

      foreach $oidName (keys %myData) {
        if ($myData{$oidName}{$entityKey} eq $entity &&
            defined $myData{$oidName}{$memberKey} &&
            defined $myData{$oidName}{$memberKey}{$currentInstanceKey}) {
          if ((defined $lookField &&
               $myData{$oidName}{$lookField} eq $lookValue) ||
              ((not defined $lookField) &&
               ($myData{$oidName}{$accumulatingKey} eq "N" ||
                ($myData{$oidName}{$accumulatingKey} eq "Y" &&
                 $myData{$oidName}{$memberKey}{$currentInstanceKey}{$resetKey} ne $resetNoRule &&
                 $myData{$oidName}{$memberKey}{$currentInstanceKey}{$resetKey} ne $resetUnknown)))) {
            if ($myData{$oidName}{$dataTypeKey} eq $timeType &&
                defined $myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey} &&
                $myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey} ne "") {
              if ($myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey} !~ /$timeLiteral/) {
                $outputValue = $myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey};

                print STDOUT "Invalid $timeType in $currValueKey ($myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey}) for instance $currentInstanceKey of $oidName.  The value will send to Prospect but no conversion to minute value will be performed.\n";
              }
              else {
                $outputValue = timeToNumber($myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey});
              }
            }
            else {
              $outputValue = $myData{$oidName}{$memberKey}{$currentInstanceKey}{$currValueKey};
            }

            printf outputFileHandle "%s|%s|%s|%s\n",
                   $myData{$oidName}{$oidKey},
                   $oidName,
                   $outputValue,
                   epochToDate($myData{$oidName}{$memberKey}{$currentInstanceKey}{$currTstampKey});
          }
        }
      }
      print outputFileHandle "END\n";
    }
  }
  close (outputFileHandle);
  rename($outputFile,  $realOutputFile) || die "Cannot rename $outputFile: $!";
}

sub gzSysFunction {
  my ($command, $fileName);
  my ($rc, @args);

  ($command, $fileName) = @_;

  @args = ($command, $fileName);

  $rc = 0xffff & system @args;

  if ($rc != 0) {
    terminalError("Could not $command $fileName");
  }
}

sub usage {
  my $message = $_[0];
  my $progName = $0;

  $progName = `basename $0`;
  chomp $progName;

  print STDERR "\nError: $message\n\n";
  print STDERR "Usage: $progName <config filename> <DC filename> <preceding timestamp> <preceding status> <DC interval> <report interval> <outputfile format>\n\n";
  print STDERR "Arguments:\n";
  print STDERR "  <config filename>     OID config filename\n";
  print STDERR "  <DC filename>         Current data collection output filename\n";
  print STDERR "  <preceding timestamp> Preceding collection timestamp in YYYYMMDDHHMI format\n";
  print STDERR "  <preceding status>    Preceding collection status\n";
  print STDERR "                        Status value\n";
  print STDERR "                          's' = successful completion\n";
  print STDERR "                          'x' = processing failed\n";
  print STDERR "                          'u' = status not available\n";
  print STDERR "  <DC interval>         Standard data collection interval in minutes\n";
  print STDERR "  <report interval>     Standard prospect reporting interval in minutes\n\n";
  print STDERR "  <outputfile format>   Output file format prefix:suffix of output file format. Example: CHICO.2:AAApm\n\n";
  exit(1);
}

sub terminalError {
  my $message = $_[0];

  print STDERR "\nError. $message.\n\n";
  exit(1);
}

sub parseTimeStamp {

	my $timeStampString = $_[0];
	my @timeStampArray = split(/\|/,$timeStampString);
	my $timeString = $timeStampArray[1];
	#This function is for converting YYYYMMDDHHMMSS to epoch time
	my $specificTimeLiteral = '\d{14}';

	my $year=0;
	my $month=0;
	my $day=0;
	my $hour=0;
	my $minute=0;
	my $sec=0;
	if ($timeString	=~ /$specificTimeLiteral/ )
	{
		$year=substr($timeString,0,4)-1900;
		$month=substr($timeString,4,2)-1;
		$day=substr($timeString,6,2);
		$hour=substr($timeString,8,2);
		$minute=substr($timeString,10,2);
		$sec=substr($timeString,12,2);

		return(timelocal($sec,$minute,$hour,$day,$month,$year));
		
	}
	return(0);	
}
